#!/usr/bin/env python

"""
TODO.TXT Manager
Adapted from:

  - todo.py by: Shane Koster <shane.koster@gmail.com>

  - todotxt concept by: Gina Trapani (ginatrapani@gmail.com)

More info:	http://todotxt.com
"""
__version__   = "1.8.1-py-beta"
__revision__  = "$Revision$"
__date__      = "2008/02/03"
__author__    = "Thibauld Nion"
__copyright__ = "Copyright 2006, Gina Trapani"
__license__   = "GPL"
__history__   = "See http://todotxt.googlecode.com/svn/trunk/CHANGELOG"



import sys
import os
from shutil import copyfile
import re
import time

try:
	import todo_birdseye as birdseye
	HAVE_BIRDSEYE=True
except:
	HAVE_BIRDSEYE=False

import todotxt.TodoTask as task
from todotxt.TodoTask import TodoTask

#--- Management of the todo file

class TodoManager(object):
	"""
	Grant access to todo files and related operations. Makes it
	possible to define a unique path the file and change this path in
	a proper way.

	Follows Borg pattern from::

	  http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66531
	"""
	__shared_state = {
		"TODO_DIR" : os.path.expanduser("~/todo"),
		# Your todo/ files locations
		"TODO_FILE"		: "",
		"DONE_FILE"		: "",
		"TODO_BACKUP"	: "",
		"DONE_BACKUP"	: "",		
		"REPORT_FILE"	: "",
		"TMP_FILE" 		: "",
		"singleLetterPriority": False,
		}
	def __init__(self):
		self.__dict__ = self.__shared_state
		if self.TODO_FILE == "":
			self._set_todo_filepaths()
	
	#--- Changing the path to the todo files

	def _set_todo_filepaths(self):
		"""
		Set the todo files paths to the right... paths
		"""
		self.TODO_FILE	= os.path.join(self.TODO_DIR,
									   "todo.txt")
		self.DONE_FILE	= os.path.join(self.TODO_DIR,
									   "done.txt")
		self.REPORT_FILE= os.path.join(self.TODO_DIR,
									   "report.txt")
		self.TMP_FILE	= os.path.join(self.TODO_DIR,
									   "todo.tmp")
		self.TODO_BACKUP= os.path.join(self.TODO_DIR,
									   "todo.bak")
		self.DONE_BACKUP= os.path.join(self.TODO_DIR,
									   "done.bak")
	
	def setTodoDir(self,new_path):
		"""
		Change the path where to look for the todo file.
		
		If the new path does not contain any existing todo.txt, then
		the former directory is entirely copied to the new location,
		thus making sure that old tasks are not lost.
		"""
		# Check if the former path contained anything
		try:
			if os.path.isfile(self.TODO_FILE):
				# Try to move the old todotxt dir to the new location.
				if not os.path.isfile(os.path.join(new_path,"todo.txt")):
					os.rename(self.TODO_DIR,new_path)
			# Enforce the name change
			self.TODO_DIR = new_path
			self._set_todo_filepaths()	
		except:
			# Put a specific message in the exception
			raise IOError("Can't set todo directory to: %s" % new_path)


	#--- Chaging global options
	
	def useSingleLetterPriority(self,choice=True):
		"""
		Tell wether the user wants to use a single capital letter (not
		enclosed in '()') to indicate the priority.
		"""
		self.singleLetterPriority = choice


	#--- Basic IO operations

	def _backup(self,orig,backup):
		"""
		Make a copy a given file.
		
		To be used before writing data.
		"""
		if not os.path.isfile(orig): return
		try:
			copyfile(orig, backup)
		except (IOError, os.error), why:
			print "Can't copy %s to %s: %s" % (orig, backup, str(why))
	
	def _backup_todo_file(self):
		"""
		Backup the todo file
		"""
		self._backup(self.TODO_FILE,self.TODO_BACKUP)

	def _backup_done_file(self):
		"""
		Backup the done file
		"""
		self._backup(self.DONE_FILE,self.DONE_BACKUP)

	def _open_todo_file(self,mode="r"):
		"""
		Open the todo file and return a pointer to it.
		Remember to close it !
		
		@param mode the file mode is the same as in the standard
		'open' function.
		"""
		if not os.path.isdir(self.TODO_DIR):
			os.mkdir(self.TODO_DIR)
		return open(self.TODO_FILE,mode)

	def _open_done_file(self,mode="r"):
		"""
		Open the done file and return a pointer to it.
		Remember to close it !
		
		@param mode the file mode is the same as in the standard
		'open' function.
		"""
		if not os.path.isdir(self.TODO_DIR):
			os.mkdir(self.TODO_DIR)
		return open(self.DONE_FILE,mode)

	def getTaskList(self):
	 	"""
	 	A utility method to obtain the list of tasks.
	 	Each task is a C{TodoTask} instance.
	 	"""
	 	tasks = []
		f = self._open_todo_file()
	 	# build a dictionary of the todo list items
	 	for count,line in enumerate(f.readlines()):
	 		if (line.strip() == ""): continue
	 		tasks.append(TodoTask(line.rstrip(),count))
		f.close()
	 	return tasks

	def getTask(self,line):
	 	"""
		Return a task corresponding to a given line in the todo file.
	 	"""
		tasks = self.getTaskList()
		if line>=len(tasks):
			raise Exception("Task not found")
		return tasks[line]
	 
	def getDoneList(self):
	 	"""
		Return a list of tasks from the DONE file.
	 	Each task is a C{TodoTask} instance.
		"""
	 	tasks = []
	 	# build a dictionary of the todo list items
		f = self._open_done_file()
	 	for count,line in enumerate(f.readlines()):
	 		if (line.strip() == ""): continue
	 		tasks.append(TodoTask(line.rstrip(),count))
		f.close()
	 	return tasks
	 
	def writeTaskList(self,tasklist):
	 	"""
	 	Write a list of TodoTask's to the TODO file
	 	"""
		self._backup_todo_file()
		f = self._open_todo_file("w")
	 	for t in sorted(tasklist,cmp=lambda x,y: cmp(x.line,y.line)):
	 		f.write(t.description + os.linesep)
	 	f.close()
	 
	def writeDoneList(self,doneList):
	 	"""
	 	Append a list of achieved tasks to the DONE file.
	 	"""
		self._backup_done_file()
		f = self._open_done_file("a")
	 	for t in doneList:
	 		f.write(t.description + os.linesep)
	 	f.close()

	#--- Manipulate the task in the files

	def checkPriority(self,text):
		"""
		Handle priority if exisiting in supplied text.
		
		Make sure that priorities entered in the following accepted
		ways, gets propermy formated in a unique (standard) way::

		[A] my important task or
		{A} my important task or
		(A) my important task

		The latter being the standard way.

		If  self.singleLetterPriority is True then you additionally can
		enter your priority as a single capital letter at the beginning
		of your todo like this::

		   A this is my highest priority

		This todo will automatically converted to::

		  (A) this is my highest priority
		"""
		if self.singleLetterPriority:
		 	re_pri = re.compile(r"^[A-Z] ")
		 	if re.search(re_pri, text):
				# A single capital letter was found, convert it to a
				# priority!
		 		newpriority = "(" + text[0] + ") "
		 		text = re.sub(re_pri, newpriority, text)
		re_pri = re.compile(r"^[\(\[\{]([a-zA-Z])[\]\}\)]") # [a] or {a} or ()
		prio = re.search(re_pri, text)
		if prio:
		 	newpriority = "(" + prio.group(1).upper() + ")"
		 	text = re.sub(re_pri, newpriority, text)
		return text

	def add(self,text):
	 	"""
		Add a new task to the TODO file.
		
		@param text the description of the task

		See also checkPriority method that is internally used.
		"""
		# Make sure the priority is correctly formated.
		text = self.checkPriority(text)
	 	f = self._open_todo_file('a')
	 	f.write(text + os.linesep)
	 	f.close()
	 
	def do(self,lines,comments=None):
	 	""" 
		Mark as done a precise task.
		
		@param lines a list of lines where to find done tasks.'
		"""
	 	tasks = self.getTaskList()
		for line in lines:
			if len(tasks) <= line:
				raise Exception("%d: No such todo."%line)
			tasks[line].do(comments)
	 	self.writeTaskList(tasks)

	def directly_done(self,description):
	 	""" 
		Add a completed task directly to done file
		"""
		t = TodoTask(description,-1)
		t.do()
		f = self._open_done_file("a")
		f.write(t.description + os.linesep)
		f.close()
	 
	def suppress(self,line):
	 	tasks = self.getTaskList()
	 	if len(tasks) <= line:
	 		raise Exception("%d: No such todo."%line)
	 	deleted = tasks.pop(line)
	 	self.writeTaskList(tasks)
		return deleted
	 
	def replace(self, line, new_text):
	 	"""replace text to a given task"""
	 	tasks = self.getTaskList()
	 	if len(tasks) <= line:
	 		raise Exception("%d: No such todo."%line)
		# make sure the prioty format is right
		new_text = self.checkPriority(new_text)
	 	tasks[line].description = new_text
	 	self.writeTaskList(tasks)

	def append_description(self,line,text):
		"""
		Append some text to the description.
		"""
		tasks = self.getTaskList()
	 	if len(tasks) <= line:
	 		raise Exception("%d: No such todo."%line)
		tasks[line].append(text)
		self.writeTaskList(tasks)

	def prepend_description(self,line,text):
		"""
		Prepend some text to the description.
		"""
		tasks = self.getTaskList()
	 	if len(tasks) <= line:
	 		raise Exception("%d: No such todo."%line)
		tasks[line].prepend(text)
		self.writeTaskList(tasks)

	def prioritize(self,line,newpriority):
		"""
		Change the priority of an item.
		"""
	 	tasks = self.getTaskList()
	 	if len(tasks) <= line:
	 		raise Exception("%d: No such todo."%line)
		tasks[line].setPriority(newpriority)
	 	self.writeTaskList(tasks)

	def removeDuplicates(self):
		"""
		Remove duplicates among todo tasks.
		"""
	 	tasks = self.getTaskList()
		dupCount = task.remove_duplicates(tasks)
	 	self.writeTaskList(tasks)
		return dupCount

	def cleanup(self,rmdup=True):
	 	"""
	 	Cleanup the tasks:
		  - archive done tasks
		  - remove duplicates

		@param rmdup if false don't the duplicates.
	 	"""
	 	tasks = self.getTaskList()
	 	donelist = []
	 	task.archive_done_tasks(tasks,donelist)
	 	self.writeDoneList(donelist)
	 	self.writeTaskList(tasks)

	#--- Gather global info

	def listKeywords(self):
		"""
		Count keywords in todo and done files
		"""
		items = []
		tasks = self.getTaskList()
		numTasks = len(tasks)

		# Add done dictionary to existing tasks
		# Can this be cone with zip() ?
		# yep with the extend method of a list
		done	 = self.getDoneList()
		tasks.extend(done)

		projects = {}
		contexts = {}

		for k,v in enumerate(tasks):
			priority = v.getPriority()
			words = v.description.split()
			for word in words:
				if len(word) < 2: continue
				if word[0:2] in ["p:", "p-"] or word.startswith("+"):
					if not projects.has_key(word):
						projects[word] = dict(ntask=0,ndone=0,tlist=[],dlist=[],priority=0)
					if k <= numTasks:
						projects[word]['ntask'] += 1
						projects[word]['tlist'].append(k)
						if priority: projects[word]['priority'] += 1
					else:
						projects[word]['ndone'] += 1
						projects[word]['dlist'].append(k)

				if word[0:1] == "@":
					if not contexts.has_key(word):
						contexts[word] = dict(ntask=0,ndone=0,tlist=[],dlist=[],priority=0)
					if k <= numTasks:
						contexts[word]['ntask'] += 1
						contexts[word]['tlist'].append(k)
						if priority: contexts[word]['priority'] += 1
					else:
						contexts[word]['ndone'] += 1
						contexts[word]['dlist'].append(k)
		return projects, contexts



	def report(self):
		"""
		Report open and closed task.
		
		Append the information to the report file.
		"""
		self.cleanup()
		active = self.getTaskList()
		closed = self.getDoneList()
		date = time.strftime("%Y-%m-%d-%T", time.localtime())
		f = open(self.REPORT_FILE, 'a')
		string = "%s %d %d" % (date, len(active), len(closed))
		f.write(string + os.linesep)
		f.close()


if __name__=="__main__":
	# Just to test the 'Borg' pattern
	t = TodoManager()
	print t.TODO_FILE

	t2 = TodoManager()
	t2.changeTodoDir("bla")
	print t.TODO_FILE
