#!/usr/bin/env python

"""
Task representation and manipulation.

Most functions adapter from the work of 
  - Shane Koster <shane.koster@gmail.com>
  - Graham Davies <grahamdaviez@gmail.com>
  - Gina Trapani (ginatrapani@gmail.com)

More info		: http://todotxt.com
"""
__author__="Thibauld Nion"
__license__="GPL, http://www.gnu.org/copyleft/gpl.html"

import time
import re
import sets

class TodoTask(object):
	"""
	Hold the text description of a task along with some metadata.
	"""
	
	def __init__(self,task_txt, line_nb): 
		"""
		Create a task with the text description and the line number.
		
		Initialise some basic metadata too.
		"""
		self.description = task_txt
		self.line = line_nb
#		self.priority = priority
#		self.project = project
#		self.context = context
#		self.done = done

	def __str__(self):
		"""
		Print the content of a task
		"""
		return self.description

	def getPriority(self):
		"""
		Return the task's priority if it has one, None elseway.
		"""
		re_pri = re.compile(r"^\(([A-Z])\) ")
		prio = re.search(re_pri, self.description)
		if prio:
			return prio.group(1)
		else:
			return None


	def setPriority(self,newpriority):
		"""
		Chagne the priority of a task.
		
		"" -> remove former priority
		"""
		if (newpriority != "" and not re.match("[A-Z]", newpriority)):
			raise Exception("Priority not recognised: %s" % newpriority)
		
		re_pri = re.compile(r"\([A-Z]\) ")

		if (newpriority == ""):
			# remove the existing priority
			self.description = re.sub(re_pri, "", self.description)
		elif (re.match(re_pri, self.description)):
			self.description = re.sub(re_pri, "(" + newpriority + ") ", self.description)
		else:
			self.description = "(" + newpriority + ") " + self.description
	
	def isDone(self):
		"""
		Check if an item is a 'done' item
		"""
		if self.description.startswith("x "): 
			return True
		else: 
			return False
	
	def do(self,comments=None):
		"""
		Mark an item as being processed or done, following the todotxt
		format: prepend 'x {current date}'.
		"""
		# done items shouldn't be done again
		if self.isDone(): return
		re_pri = re.compile(r"^\([A-Z]\) ")
		self.description = re.sub(re_pri, "", self.description)

		# add comment text if any
		if comments: 
			self.addComment(comments)
		# put the "done stamp"
		date = time.strftime("%Y-%m-%d", time.localtime())
		self.description = " ".join(["x", date,
									 self.description])	

	def append(self,text):
		"""
		Append text to the text of a given task.
		"""
		self.description += " %s"%text
	 
	def prepend(self,text):
		"""
		Prepend text to a given task.
		
		Preserve a priority at the beginning of the line, if one
		exists.
		"""
		# handle priority, if existing
		re_pri = re.compile(r"^\([A-Z]\) ")
		prio = re.search(re_pri, self.description)
		if prio:
			self.description = re.sub(re_pri, "", self.description)
			self.description = " ".join([text, self.description])
			self.description = prio.group(0) + self.description
		else:
			self.description = " ".join([text, self.description])
	
	def addComment(self,comments):
		"""
		Add a comment to the task description.
		"""
		self.description += " # %"%comments
		
	def rmExtension(self, attribute):
		"""remove an extension attribute"""
		newitem = self.description
		extstart = newitem.find(" {" + attribute)
		if extstart < 0:
			return
		extend = newitem.find("}", extstart+1)
		if extend < 0:
			return
		newitem = newitem[:extstart] + newitem[extend+1:]
		self.description = newitem

	def addExtension(self, attribute, value):
		"""append an extension attribute"""
		newextension = "{" + attribute + ": " + value + "}"
		self.append(newextension)


#--- List Manipulations


def filter_list(tasklist, patterns=None,
				escape=True, matchAny=False, complement=False):
	"""
	Filter out the tasks that do not match the patterns. if patterns
	is None, keep all the items and return the list as is.

	@param tasklist a list of todo tasks

	@param patterns regexp patterns to be mathched

	@param escape tells wether the patterns have to be escaped

	@param matchAny if true then matching only one of the patterns is
	enough to be selected.

	@param complement if true then the tasks *not matching* the
	patterns are returned.
	"""
	items = []
	if patterns:
		for v in tasklist:
			# Match any or all tasks using matchAny switch
			match = True
			if matchAny: match = False
			# Apply the patterns
			for pattern in patterns:
				if escape: pattern = re.escape(pattern)
				if matchAny:
					if (re.search(pattern, v.description, re.IGNORECASE)): match = True
				else:
					if (not re.search(pattern, v.description, re.IGNORECASE)): match = False
			# Actually select (or not...)
			if complement:
				# adds tasks if not matched
				if not match: items.append(v)
			else:
				# add if matched
				if match: items.append(v)
	else:
		# just to ease the use of the function
		items = tasklist

	return items


def dates_filter_list(tasklist,datelist):
	"""
	Match tasks with a dates in datelist.
	"""
	items = []
	for v in tasklist:
		match = False
		for date in datelist:
			if (re.search(date, v.description, re.IGNORECASE)): 
				match = True
		if (match == True): items.append(v)
	return items


def sort_list(tasklist):
	"""
	Sort the list and takes into account the explicit priorities (with
	the '(A)' format)
	"""
	re_pri = re.compile(r".*(\([A-Z]\)).*")
	pos_dict = {}
	no_prio = []
	# Gather the result is "hash table" like structures
	for pos in range(len(tasklist)):
		item = tasklist[pos]
		m = re.search(re_pri,item.description)
		if m is None:
			# when no priority is found store the result in a special
			# list.
			no_prio.append(pos)
		else:
			# get the priority label (prio= "(A)", "(B)", etc)
			prio = m.group(1)
			# if the label has not been seen yet create an entry in
			# the dict
			if not pos_dict.has_key(prio):
				pos_dict[prio] = []
			pos_dict[prio].append(pos)
	# now let do the sorting !
	prio_list = pos_dict.keys()
	prio_list.sort()
	sorted_tasks = []
	for prio in prio_list:
		for pos in pos_dict[prio]:
			t = tasklist[pos]
			sorted_tasks.append(t)
	for pos in no_prio:
		sorted_tasks.append(tasklist[pos])
	return sorted_tasks

def remove_duplicates(tasklist):
	"""
	Remove duplicate items.
	"""
	theSet = sets.Set(tasklist)
	dupCount = len(tasklist) - len(theSet)
	if dupCount > 0:
		new_list = []
		for v in tasklist:
			if v in theSet:
				new_list.append(v)
				theSet.remove(v)
	return dupCount
	
def archive_done_tasks(taskList,doneList):
	"""
	Archive done tasks by puttint them in the doneList and removing
	them from the task list.
	"""
	to_pop = []
	for k,t in enumerate(taskList):
		if t.isDone():
			doneList.append(t)
			to_pop.append(k)
	# reverse the index list to keep the indices valid untill the end.
	to_pop.reverse()
	for i in to_pop:
		taskList.pop(i)
	return True

