#!/usr/bin/env python

"""
TodoListView definition.

A widget that will display a todo list with the GTK Tree Model.
"""
__author__="Thibauld Nion"

import re
import os
import sys
from gettext import gettext as _

import gtk
import pango

# dev only !
sys.path.append("/home/thibauld/src/todobar")

from todotxt.TodoManager import TodoManager

class TodoEditTextView:
	"""
	Display a textview widget and get the text.
	"""

	def __init__(self,task_id,description=None):
		"""
		task_id corresponds to the line number of the task. If
		task_id==-1 then the task is considered as a new task.

		If description, then this is the default text that will be
		displayed in the text view.
		"""
		# create a new window
		self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.window.set_title(_("Todo ADD"))
		
		# access to the todo manager
		self._todo_manager = TodoManager()

		# Customisable variables
		# Todo: put them in gconf
		self.TXT_WRAP_LENGTH = 80
		self.task_id = task_id
		
		instructions = gtk.Label(_("Task's description."))

		# Create the view
		self._init_text_view(description)
		# Put it in a scroll bar
		self.scrolledwindow = gtk.ScrolledWindow()
		self.scrolledwindow.set_policy(gtk.POLICY_AUTOMATIC,
									   gtk.POLICY_AUTOMATIC)
		self.scrolledwindow.add(self.textview)

		# Create buttons
		self._init_bottom_buttons()
	
		# This packs the buttons and the treeview.
		self.mainBox = gtk.VBox()
		self.mainBox.pack_start(instructions)
		self.mainBox.pack_start(self.scrolledwindow,expand=True)
		self.mainBox.pack_start(self.downBox,expand=False)
		self.window.add(self.mainBox)

		# When the window is given the "delete_event" signal
		self.window.connect("delete_event", self.delete_event)
		# Here we connect the "destroy" event 
		self.window.connect("destroy", self.destroy)

		# Ask for a good initial size
		self._set_good_window_size() 
		# The final step is to display this newly created window.
		self.window.show_all()


	def _set_good_window_size(self):
		"""
		Try to set a good initial size for the window
		"""
		pctx = self.textview.get_pango_context()
		l = pango.Layout(pctx)
		sampletxt = ["a" for i in range(self.TXT_WRAP_LENGTH)]
		l.set_text("".join(sampletxt))
		txt_width, txt_height = l.get_pixel_size()
		# get a little more space than needed
		width = min(int(txt_width),
					int(gtk.gdk.screen_width()*2./3.))
		height = min(int(3*txt_height),
					 int(gtk.gdk.screen_height()*.25))
		self.textview.set_size_request(width,height)
		

	def _init_bottom_buttons(self):
		"""
		Layout the bottom buttons
		"""		
		# Creates validation/cancellation buttons.
		btnCancel = gtk.Button(stock=gtk.STOCK_CANCEL)
		if self.task_id<0:
			btnApply = gtk.Button(stock=gtk.STOCK_ADD)
			btnApply.connect("clicked", self.add_task, None)
		else:
			btnApply = gtk.Button(stock=gtk.STOCK_APPLY)
			btnApply.connect("clicked", self.edit_task, None)
		# connect events
		btnCancel.connect("clicked", self.cancel, None)
		# Layout the lower buttons
		self.downBox = gtk.HBox()
		self.downBox.pack_start(btnCancel)
		self.downBox.pack_end(btnApply)



	def _init_text_view(self,description):
		"""
		Create the text view that where the user will enter the task.
		"""		
		self.textview = gtk.TextView()
		self.textbuffer = self.textview.get_buffer()
		# Set some properties
		self.textview.set_editable(True)
		self.textview.set_wrap_mode(gtk.WRAP_WORD)
		self.textview.set_accepts_tab(False)
		self.textview.connect("key-press-event",self.no_return)
		if description is not None:
			self.textbuffer.set_text(description)

	def no_return(self, widget, event):
		"""
		When text is inserted make sure that we wont accept the return
		key.
		"""
		# test if we the return key has been pressed 
		if 65293 != event.keyval:
			return False
		else:
			return True

	def add_task(self,widget,data=None):
		"""
		Add the task to the list
		"""
		startiter, enditer = self.textbuffer.get_bounds()
		txt = self.textbuffer.get_text(startiter,enditer)
		self._todo_manager.add(txt)
		self.window.destroy()

	def edit_task(self,widget,data=None):
		"""
		Apply and save the modifications of the tasks
		"""
		startiter, enditer = self.textbuffer.get_bounds()
		txt = self.textbuffer.get_text(startiter,enditer)
		self._todo_manager.replace(self.task_id,txt)
		self.window.destroy()
		
	def cancel(self,widget,data=None):
		"""
		Do not apply the done/not done settings
		"""
		self.window.destroy()

	def delete_event(self, widget, event, data=None):
		"""
		Change FALSE to TRUE and the main window will not be destroyed
		with a "delete_event".
		"""
		return False

	def destroy(self, widget, data=None):
		pass


class TodoAddTextView(TodoEditTextView):
	"""
	Display a textview widget and get the text.
	
	Specific for a new task to add to the list.
	"""

	def __init__(self,description=None):
		TodoEditTextView.__init__(self,-1,description)


if __name__ == "__main__":
	tdlv = TodoAddTextView()
	# All PyGTK applications must have a gtk.main(). Control ends here
	# and waits for an event to occur (like a key press or mouse event).
	gtk.main()

