#!/usr/bin/env python

"""
TodoListView definition.

A widget that will display a todo list with the GTK Tree Model.
"""
__author__="Thibauld Nion"

import re
import os
import sys
import time
from gettext import gettext as _
from textwrap import fill as txt_fill

import gtk
import pango

# dev only !
sys.path.append("/home/thibauld/src/todobar")

from todotxt.TodoManager import TodoManager
import todotxt.TodoTask as tasklib


class TodoListView:
	"""
	Display a todo list and let the user check the 'done' items.
	"""
	done_icons = {
		True: gtk.STOCK_REMOVE,
		False: gtk.STOCK_GO_FORWARD,
		}


	def __init__(self,tasklist):
		# create a new window
		self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.window.set_title(_("Todo LIST"))

		# Customisable variables
		# Todo: put them in gconf
		self.TXT_WRAP_LENGTH = 80
		self.AUTO_ARCHIVE = True

		# access to the todo manager
		self._todo_manager = TodoManager()

		# Create a list store model
		self._init_store(tasklist)
		# Create the view
		self._init_tree_view()
		# Put it in a scroll bar
		self.scrolledwindow = gtk.ScrolledWindow()
		self.scrolledwindow.set_policy(gtk.POLICY_AUTOMATIC,
									   gtk.POLICY_AUTOMATIC)
		self.scrolledwindow.add(self.treeview)

		# Create buttons
		self._init_bottom_buttons()
	
		# This packs the buttons and the treeview.
		self.mainBox = gtk.VBox()
		self.mainBox.pack_start(self.scrolledwindow,expand=True)
		self.mainBox.pack_start(self.downBox,expand=False)
		self.window.add(self.mainBox)

		# When the window is given the "delete_event" signal
		self.window.connect("delete_event", self.delete_event)
		# Here we connect the "destroy" event 
		self.window.connect("destroy", self.destroy)

		# Ask for a good initial size
		self._set_good_window_size(len(tasklist)) 
		# The final step is to display this newly created window.
		self.window.show_all()

	def _init_store(self,tasklist):
		"""
		Create the liststore model.
		"""
		# In the model we have:
		# [0]: status (done,todo)
		# [1]: funny icon (linked to status)
		# [2]: text description
		# [3]: original line number of task
		# [4]: boolean indicating the initial status 
		# was actually changed
		self.liststore = gtk.ListStore(bool, str, str, int, bool)
		# Fill in the list store
		for t in tasklist:
			self.liststore.append([t.isDone(), 
								   self.done_icons[t.isDone()],
								   txt_fill(t.description,
											self.TXT_WRAP_LENGTH),
								   t.line, False])

	def _set_good_window_size(self,nb_items):
		"""
		Try to set a good initial size for the window
		"""
		pctx = self.treeview.get_pango_context()
		l = pango.Layout(pctx)
		l.set_text("".join(["a" for i in range(self.TXT_WRAP_LENGTH)]))
		txt_width, txt_height = l.get_pixel_size()
		# get a little more space for the check box and icon
		width = min(int(txt_width),
					int(gtk.gdk.screen_width()*2./3.) )
		# approximate the vertical height by considering that some
		# tasks have to be put on 2 or more lines
		height = min( int(3 * nb_items * txt_height),
					  int(gtk.gdk.screen_height()*2./3.) )
		self.treeview.set_size_request(width,height)
		

	def _init_bottom_buttons(self):
		"""
		Layout the bottom buttons
		"""
		
		# Creates validation/cancellation buttons.
		btnCancel = gtk.Button(stock=gtk.STOCK_CANCEL)
		btnApply = gtk.Button(stock=gtk.STOCK_APPLY)
		# connect events
		btnCancel.connect("clicked", self.cancel, None)
		btnApply.connect("clicked", self.apply, None)
		# Layout the lower buttons
		self.downBox = gtk.HBox()
		self.downBox.pack_start(btnCancel)
		self.downBox.pack_end(btnApply)



	def _init_tree_view(self):
		"""
		Create the tree view that will display the liststore.
		"""		
		self.treeview = gtk.TreeView(self.liststore)
		# Set some properties
		self.treeview.set_enable_search(True)
		self.treeview.set_rules_hint(True)
		self.treeview.set_search_column(1)

		# create a CellRenderers to render the data
		cellpb = gtk.CellRendererPixbuf()
		cellchk = gtk.CellRendererToggle()
		cell = gtk.CellRendererText()

		cellchk.set_property('activatable', True)
		cellchk.connect("toggled",self.task_toggled,self.liststore)
		
		# Create the TreeViewColumns to display the data
		tvcolumn = gtk.TreeViewColumn(_('Status'),cellchk)
		tvcolumn1 = gtk.TreeViewColumn(_('Task description'))
		tvcolumn1.pack_start(cellpb, expand=False)
		tvcolumn1.pack_start(cell, expand=True)

		# Add columns to treeview
		self.treeview.append_column(tvcolumn)
		self.treeview.append_column(tvcolumn1)
		# Link the content of the columns with the content of the
		# store
		tvcolumn.add_attribute(cellchk, "active", 0)
		tvcolumn1.set_attributes(cellpb, stock_id=1)
		tvcolumn1.set_attributes(cell, markup=2)


	def task_toggled(self,cell, path, model):
		"""
		When user declare a task as done or not done after all.
		"""
		model[path][0] = not model[path][0]
		model[path][1] = self.done_icons[model[path][0]]		
		# automatically decides if the status has been changed or if
		# the change has been undone:
		model[path][4] = not model[path][4]
# 		print "Toggle '%s' to: %s" % (model[path][0], 
# 									  model[path][1])
		return

	def apply(self,widget,data=None):
		"""
		Apply and save the modifications of the tasks
		"""
		tasks = self._todo_manager.getTaskList()
		for row in self.liststore:
			if row[4]:# if the task's status has changed
				tasks[row[3]].do()
		donelist = []
		if self.AUTO_ARCHIVE:
			tasklib.archive_done_tasks(tasks,donelist)
		self._todo_manager.writeDoneList(donelist)
		self._todo_manager.writeTaskList(tasks)
		self.window.destroy()


	def cancel(self,widget,data=None):
		"""
		Do not apply the done/not done settings
		"""
		self.window.destroy()

	def delete_event(self, widget, event, data=None):
		"""
		Change FALSE to TRUE and the main window will not be destroyed
		with a "delete_event".
		"""
		return False

	def destroy(self, widget, data=None):
#		gtk.main_quit()
		pass


if __name__ == "__main__":
	t = getTaskList()
	t = sort_list(t)
	tdlv = TodoListView(t)
	# All PyGTK applications must have a gtk.main(). Control ends here
	# and waits for an event to occur (like a key press or mouse event).
	gtk.main()

