#!/usr/bin/env python

"""
TODO.TXT Manager
Author:	 Thibauld Nion <tibonihoo _at_ yahoo (dot) fr>
Adapted from 
todo.py by: Shane Koster <shane.koster@gmail.com>
Concept by: Gina Trapani (ginatrapani@gmail.com)
License:	GPL, http://www.gnu.org/copyleft/gpl.html
Version:	1.5.2-py
More info:	http://todotxt.com
"""

import re
import os
import sys
import time
try:
	import todo_birdseye as birdseye
	HAVE_BIRDSEYE=True
except:
	HAVE_BIRDSEYE=False


TODO_DIR = os.path.expanduser("~/todo")
TODO_LANG_DIR = os.path.join(os.path.dirname(__file__),"lang")

# Your todo/done/report.txt locations
TODO_FILE	= TODO_DIR + "/todo.txt"
DONE_FILE	= TODO_DIR + "/done.txt"
REPORT_FILE = TODO_DIR + "/report.txt"
TMP_FILE	= TODO_DIR + "/todo.tmp"


class TodoTask(object):
	"""
	Hold the text description of a task along with some metadata.
	"""
	
	def __init__(self,task_txt, line_nb, 
				 priority=None, done=False): 
#				 context=None, project=None):
		"""
		Create a task with the text description and the line number.
		
		Initialise some basic metadata too.
		"""
		self.description = task_txt
		self.line = line_nb
		self.priority = priority
#		self.project = project
#		self.context = context
		self.done = done



#--- Access the todo file (I/O)

def add(text):
	"""add a new task to the TODO file"""
	f = open(TODO_FILE, "a")
	f.write(text + os.linesep)
	f.close()

def done(line):
	""" mark as done a precise task (identified by its line number)"""
	tasks = getTaskList()
	if len(tasks) <= line:
		raise Exception("No such task: #%s"%line)
	mark_as_done(tasks[line])
	writeTaskList(tasks)

def suppress(line):
	tasks = getTaskList()
	if len(tasks) <= line:
		raise Exception("No such task: #%s"%line)
	tasks.pop(line)
	writeTaskList(tasks)

def edit(line, new_text):
	"""replace text to a given task"""
	tasks = getTaskList()
	if len(tasks) <= line:
		raise Exception("No such task: #%s"%line)
	tasks[line].description = new_text
	writeTaskList(tasks)
	
# utilities manipulating lists of TodoTask instances

def getTaskList():
	"""
	A utility method to obtain a list 
	of tasks from the TODO file.

	Each task is itself a list containing two elements (the task's
	text description and the task's line number). 

	Each post processing function can then append it's own type of
	elements to the list. YES, is the object of the poor, but it's
	still handy here.
	"""
	tasks = []
	# build a dictionary of the todo list items
	for count,line in enumerate(open(TODO_FILE).readlines()):
		if (line.strip() == ""): continue
		tasks.append(TodoTask(line.rstrip(),count))
	return tasks

def getDoneList():
	"""Return a list of tasks from the DONE file"""
	tasks = []
	# build a dictionary of the todo list items
	for count,line in enumerate(open(DONE_FILE).readlines()):
		if (line.strip() == ""): continue
		tasks.append(line.rstrip(),count)
	return tasks

def writeTaskList(tasklist):
	"""
	Write a list of TodoTask's to the TODO file
	"""
	f = open(TODO_FILE, "w")
	for t in sorted(tasklist,cmp=lambda x,y: cmp(x.line,y.line)):
		f.write(t.description + os.linesep)
	f.close()

def writeDoneList(doneList):
	"""
	Append a list of achieved tasks to the DONE file.
	"""
	f = open(DONE_FILE, "a")
	for t in doneList:
		f.write(t.description + os.linesep)
	f.close()

# utilities for	 dictionnaries filled with texts

def getTaskDict():
	"""a utility method to obtain a dictionary of tasks from the TODO file"""
	count = 0
	tasks = {}
	# build a dictionary of the todo list items
	for line in open(TODO_FILE).readlines():
		if (line.strip() == ""): continue
		count = count + 1
		tasks[count] = line.rstrip()
	return tasks
	

def getDoneDict():
	"""a utility method to obtain a dictionary of tasks from the DONE file"""
	count = 0
	tasks = {}
	# build a dictionary of the todo list items
	for line in open(DONE_FILE).readlines():
		if (line.strip() == ""): continue
		count = count + 1
		tasks[count] = line.rstrip()
	return tasks
	

def writeTasksDict(taskDict):
	"""a utility method to write a dictionary of tasks to the TODO file"""
	keys = taskDict.keys()
	keys.sort()
	f = open(TODO_FILE, "w")
	for key in keys:
		f.write(taskDict[key] + os.linesep)
	f.close()
	
def writeDone(doneDict):
	keys = doneDict.keys()
	keys.sort()
	f = open(DONE_FILE, "a")
	for key in keys:
		f.write(doneDict[key] + os.linesep)
	f.close()


#--- Manipulate the task list

def filter_list(tasklist, patterns=None):
	"""
	Filter out the tasks that do not match the patterns. if patterns
	is None, keep all the items and return the list as is.
	"""
	items = []

	if patterns is not None:
		for v in tasklist:
			match = True
			# make sure all patterns are matched
			for pattern in patterns:
				if (not re.search(pattern, v.description, re.IGNORECASE)): 
					match = False

			if (match == True): 
				items.append(v)
	else:
		items = tasklist

	return items

def sort_list(tasklist):
	"""
	Sort the list and takes into account the explicit priorities (with
	the '(A)' format)
	"""
	re_pri = re.compile(r".*(\([A-Z]\)).*")
	pos_dict = {}
	no_prio = []
	# Gather the result is "hash table" like structures
	for pos in range(len(tasklist)):
		item = tasklist[pos]
		m = re.search(re_pri,item.description)
		if m is None:
			# when no priority is found store the result in a special
			# list.
			no_prio.append(pos)
		else:
			# get the priority label (prio= "(A)", "(B)", etc)
			prio = m.group(1)
			# if the label has not been seen yet create an entry in
			# the dict
			if not pos_dict.has_key(prio):
				pos_dict[prio] = []
			pos_dict[prio].append(pos)
	# now let do the sorting !
	prio_list = pos_dict.keys()
	prio_list.sort()
	sorted_tasks = []
	for prio in prio_list:
		for pos in pos_dict[prio]:
			t = tasklist[pos]
			t.priority = prio
			sorted_tasks.append(t)
	for pos in no_prio:
		sorted_tasks.append(tasklist[pos])
	return sorted_tasks


def check_list(tasklist):
	"""
	Check wether tasks are done or not.
	"""
	re_done = re.compile('^x \d+')
	for t in tasklist:
		m = re_done.match(t.description)
		if m is None:
			t.done = False
		else:
			t.done = True
	return tasklist



def archive_done_tasks(taskList,doneList):
	"""
	Archive done tasks by puttint them in the doneList and removing
	them from the task list.
	"""
	to_pop = []
	for k,t in enumerate(taskList):
		if t.description.startswith("x "):
			doneList.append(t)
			to_pop.append(k)
	# reverse the index list to keep the indices valid untill the end.
	to_pop.reverse()
	for i in to_pop:
		taskList.pop(i)
	return True

def cleanup():
	"""
	Cleanup the tasks
	"""
	tasks = getTaskList()
	donelist = []
	archive_done_tasks(tasks,donelist)
	writeDoneList(donelist)
	writeTaskList(tasks)



#--- Modify the text description of a task

def mark_as_done(task):
	"""
	Mark an item as done, following the todotxt format (prepend an 'x'
	and current date).
	"""
	date = time.strftime("%Y-%m-%d", time.localtime())
	task.description = " ".join(["x", date,
								 task.description])



# to triage


def append(item, text=""):
	"""append text to a given task"""
	tasks = getTaskDict()
	if (not tasks.has_key(item)):
		print "%d: No such todo." % item
		sys.exit(1)
	tasks[item] = " ".join([tasks[item], text])
	writeTasksDict(tasks)


def delete(item):
	tasks = getTaskDict()
	if (not tasks.has_key(item)):
		print "%d: No such todo." % item
		sys.exit(1)
	tasks.pop(item)
	writeTasksDict(tasks)

def list(patterns=None):
	items = [] 
	tasks = getTaskDict()

	if (patterns):
		for k,v in tasks.iteritems():
			match = True
			# make sure all patterns are matched
			for pattern in patterns:
				if (not re.search(pattern, v, re.IGNORECASE)): match = False

			if (match == True): items.append("%3d: %s" % (k, v))
	else:
		for k,v in tasks.iteritems():
			items.append("%3d: %s" % (k, v)) 

	#items.sort() # sort by todo.txt order
	items.sort(alphaSort) # sort by tasks alphbetically

	re_pri = re.compile(r".*(\([A-Z]\)).*")
	for item in items: 
		print re_pri.sub(highlightPriority, item)

def alphaSort(a, b):
	"""sorting function to sort tasks alphabetically"""
	if (a[5:] > b[5:]): return 1
	elif (a[5:] < b[5:]): return -1
	else: return 0

def highlightPriority(matchobj):
	"""color replacement function used when highlighting priorities"""
	if (matchobj.group(1) == "(A)"):
		return PRI_A + matchobj.group(0) + DEFAULT
	elif (matchobj.group(1) == "(B)"):
		return PRI_B + matchobj.group(0) + DEFAULT
	elif (matchobj.group(1) == "(C)"):
		return PRI_B + matchobj.group(0) + DEFAULT
	else: 
		return PRI_X + matchobj.group(0) + DEFAULT

def prioritize(item, newpriority):
	newpriority = newpriority.upper()
	tasks = getTaskDict()
	if (not tasks.has_key(item)):
		print "%d: No such todo." % item
		sys.exit(1)
	if (newpriority != "" and not re.match("[A-Z]", newpriority)):
		print "Priority not recognized: " + newpriority
		sys.exit(1)

	re_pri = re.compile(r"\([A-Z]\) ")

	if (newpriority == ""):
		# remove the existing priority
		tasks[item] = re.sub(re_pri, "", tasks[item])
	elif (re.match(re_pri, tasks[item])):
		tasks[item] = re.sub(re_pri, "(" + newpriority + ") ", tasks[item])
	else:
		tasks[item] = "(" + newpriority + ") " + tasks[item]

	writeTasksDict(tasks)

def replace(item, text):
	"""replace text to a given task"""
	try:
		edit(item,text)
	except e:
		print e
		sys.exit(1)

def removeDuplicates():
	pass

def report():
	# archive first
	archive()

	active = getTaskDict()
	closed = getDoneDict()

	date = time.strftime("%Y-%m-%d-%T", time.localtime())
	f = open(REPORT_FILE, 'a')
	string = "%s %d %d" % (date, len(active), len(closed))
	f.write(string + os.linesep)
	f.close()

if __name__ == "__main__":
	if (len(sys.argv) < 2):
		usage()
		sys.exit()
	else:
		if (sys.argv[1] == "-nc"):
			PRI_A = NONE
			PRI_B = NONE
			PRI_C = NONE
			PRI_X = NONE
			action = sys.argv[2]
			args = sys.argv[3:]
		else:
			action = sys.argv[1]
			args = sys.argv[2:]

	if (action == "add"):
		if (args): add("".join(args))
		else: print "Usage: " + sys.argv[0] + " add TEXT [p:PROJECT] [@CONTEXT]"
	elif (action == "append"):
		if (len(args) > 1 and args[0].isdigit()): 
			append(int(args[0]), "".join(args[1:]))
		else:
			print "Usage: " + sys.argv[0] + " append <item_num> TEXT"
	elif (action == "archive"):
		archive()
	elif (action == "del"):
		if (len(args) == 1 and args[0].isdigit()):
			delete(int(args[0]))
		else:
			print "Usage: " + sys.argv[0] + " del <item_num>"
	elif (action == "do"):
		if (len(args) == 1 and args[0].isdigit()):
			do(int(args[0]))
		else:
			print "Usage: " + sys.argv[0] + " do <item_num>"
	elif (action == "ls" or action == "list"):
		if (len(args) > 0):
			list(args)
		else:
			list()
	elif (action == "lspri" or action == "listpri"):
		if (len(args) > 0): 
			x = ["\([" + args[0] + "]\)"]
		else: 
			x = ["\([A-Z]\)"]
		list(x)
	elif (action == "pri"):
		if (len(args) == 2 and args[0].isdigit() and args[1].isalpha()): 
			prioritize(int(args[0]), args[1])
		elif (len(args) == 1 and args[0].isdigit()): 
			# remove the existing priority
			prioritize(int(args[0]), "")
		else:
			print "Usage: " + sys.argv[0] + " pri <item_num> [PRIORITY]"
	elif (action == "replace"):
		if (len(args) > 2 and args[0].isdigit()):
			replace(int(args[0]), " ".join(args[1:]))
		else:
			print "Usage: " + sys.argv[0] + " replace <item_num> TEXT"
	elif (action == "remdup"):
		removeDuplicates()
	elif (action == "report"):
		report()
	elif (action == "birdseye"):
		if HAVE_BIRDSEYE:
			birdseye.main([TODO_FILE,DONE_FILE])
		else:
			print "Birdseye functionality is not available"			
	else:
		usage()
